--					FUNCTIONS					--

--Acquires and validates the active sprite
local function acquire_sprite()
	local sprite = app.sprite
	if sprite == nil then
		app.alert{ title = "Error", text = "No sprite was selected" }
		return nil
	end
	if sprite.colorMode ~= ColorMode.INDEXED then
		app.alert{ title = "Error", text = "Sprite colours must be indexed" }
		return nil
	end
	return sprite
end

--Acquires a list of all the usable pixel layers
local function acquire_layers(sprite)
	local layers = { }
	for layer_index = 1, #sprite.layers do
		local image = sprite.layers[layer_index]
		if image.isImage then
			layers[#layers + 1] = image
		end
	end
	return layers
end

--Creates and sets up the custom export dialog
local function create_dialog(sprite, path)
	local dialog = Dialog("Layered Mesh 3D Exporter")
	local function set_format_type()
		local dialog_data = dialog.data
		local show_embed = dialog_data.is_textured and dialog_data.format_text
		local show_files = show_embed and not dialog_data.embedded_texture
		dialog:modify{
			id = "embedded_texture",
			visible = show_embed
		}
		dialog:modify{
			id = "texture_file",
			visible = show_files
		}
		dialog:modify{
			id = "location_label",
			text = "Select a location to save the file"..(show_files and "s." or ".")
		}
		local file_name = dialog_data.mesh_file:sub(1, #dialog_data.mesh_file - #app.fs.fileExtension(dialog_data.mesh_file))
		if dialog_data.format_binary then
			dialog:modify{
				id = "mesh_file",
				filename = file_name.."glb",
				filetypes = { "glb" }
			}
		else
			dialog:modify{
				id = "mesh_file",
				filename = file_name.."gltf",
				filetypes = { "gltf" }
			}
		end
	end
	local function set_is_textured()
		local dialog_data = dialog.data
		local show_size = dialog_data.is_textured
		local show_embed = show_size and dialog_data.format_text
		local show_files = show_embed and not dialog_data.embedded_texture
		dialog:modify{
			id = "embedded_texture",
			visible = show_embed
		}
		dialog:modify{
			id = "texture_width",
			visible = show_size
		}
		dialog:modify{
			id = "texture_height",
			visible = show_size
		}
		dialog:modify{
			id = "location_label",
			text = "Select a location to save the file"..(show_files and "s." or ".")
		}
		dialog:modify{
			id = "texture_file",
			visible = show_files
		}
	end
	local function set_texture_embedded()
		local dialog_data = dialog.data
		local show_extras = dialog_data.is_textured
		local show_files = show_extras and not dialog_data.embedded_texture
		dialog:modify{
			id = "texture_width",
			visible = show_extras
		}
		dialog:modify{
			id = "texture_height",
			visible = show_extras
		}
		dialog:modify{
			id = "location_label",
			text = "Select a location to save the file"..(show_files and "s." or ".")
		}
		dialog:modify{
			id = "texture_file",
			visible = show_files
		}
	end
	local function set_scaling_mode()
		local show_factor = dialog.data.custom_scaling
		dialog:modify{
			id = "scale_x",
			visible = show_factor
		}
		dialog:modify{
			id = "scale_y",
			visible = show_factor
		}
		dialog:modify{
			id = "scale_z",
			visible = show_factor
		}
	end
	dialog:label{ text = "Basic output details." }
	dialog:entry{
		id = "name_entry",
		label = "Name",
		text = app.fs.fileName(path):gsub(app.fs.filePath(path), "")
	}
	dialog:number{
		id = "frame_from",
		label = "Frames",
		decimals = 0,
		text = "1",
	}
	dialog:number{
		id = "frame_to",
		decimals = 0,
		text = ""..#sprite.frames,
	}
	dialog:radio{
		id = "format_binary",
		label = "Format",
		text = "Binary",
		selected = true,
		onclick = set_format_type
	}
	dialog:radio{
		id = "format_text",
		text = "Text",
		selected = false,
		onclick = set_format_type
	}
	dialog:separator{ }
	dialog:label{ text = "Extra output details." }
	dialog:check{
		id = "is_textured",
		label = "Colours",
		text = "Textured",
		selected = false,
		onclick = set_is_textured
	}
	dialog:check{
		id = "embedded_texture",
		text = "Embedded",
		selected = false,
		visible = false,
		onclick = set_texture_embedded
	}
	dialog:radio{
		id = "no_normals",
		label = "Normals",
		text = "None",
		selected = true,
	}
	dialog:radio{
		id = "flat_normals",
		text = "Flat",
		selected = false
	}
	dialog:radio{
		id = "no_scaling",
		label = "Scaling",
		text = "None",
		selected = true,
		onclick = set_scaling_mode
	}
	dialog:radio{
		id = "normalised_scaling",
		text = "Normalised",
		selected = false,
		onclick = set_scaling_mode
	}
	dialog:radio{
		id = "custom_scaling",
		text = "Custom",
		selected = false,
		onclick = set_scaling_mode
	}
	dialog:number{
		id = "texture_width",
		label = "Cell Size",
		text = "32",
		decimals = 0,
		visible = false
	}
	dialog:number{
		id = "texture_height",
		text = "32",
		decimals = 0,
		visible = false
	}
	dialog:number{
		id = "scale_x",
		label = "Scale Factor",
		text = "1.0000",
		decimals = 4,
		visible = false
	}
	dialog:number{
		id = "scale_y",
		text = "1.0000",
		decimals = 4,
		visible = false
	}
	dialog:number{
		id = "scale_z",
		text = "1.0000",
		decimals = 4,
		visible = false
	}
	dialog:separator{ }
	dialog:label{
		id = "location_label",
		text = "Select a location to save the file."
	}
	dialog:file{
		id = "mesh_file",
		label = "Mesh",
		title = "Mesh File Export",
		filename = path..".glb",
		filetypes = { "glb" },
		open = false,
		save = true
	}
	dialog:file{
		id = "texture_file",
		label = "Texture",
		title = "Texture File Export",
		filename = path..".png",
		filetypes = { "png", "bmp" },
		open = false,
		save = true,
		visible = false
	}
	dialog:separator{ }
	dialog:button{
		id = "confirm",
		text = "Confirm"
	}
	dialog:button{
		id = "cancel",
		text = "Cancel"
	}
	return dialog
end

--Outputs a binary byte of the given value to append to a string
local function string_byte(value)
	return string.char(value & 0xFF)
end

--Outputs a binary word of the given value to append to a string
local function string_word(value)
	return string.char(value & 0xFF)..string.char((value >> 8) & 0xFF)
end

--Outputs a binary dword of the given value to append to a string
local function string_dword(value)
	return string.char(value & 0xFF)..string.char((value >> 8) & 0xFF)..string.char((value >> 16) & 0xFF)..string.char((value >> 24) & 0xFF)
end

--Outputs a binary float of the given value to append to a string
local function string_float(value)
	local sequence = ""
	local packed = string.pack('f', value)
	for index = 1, 4 do
		sequence = sequence..string.char(string.byte(packed, index))
	end
	return sequence
end

--Outputs a base64 encoded copy of the given data to append to a string
local function string_base64(data)
	local key = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'
    return ((data:gsub(".", function(x) 
		local result = "";
		local byte = x:byte()
		for index = 1, 8 do
			result = result..(((byte % (2 ^ index)) - (byte % (2 ^ (index - 1))) > 0) and "1" or "0")
		end
		return string.reverse(result)
	end).."0000"):gsub("%d%d%d?%d?%d?%d?", function(x)
		if (#x < 6) then return "" end
		local token = 1
		for index = 1, 6 do
			token = token + (x:sub(index, index) == "1" and 2 ^ (6 - index) or 0)
		end
		return key:sub(token, token)
	end)..(({ "", "==", "=" })[(#data % 3) + 1]))
end

--Adds the specified vertex to the given table and outputs the index
local function add_vertex(sprite, vertices, x, y, z, scaling, nx, ny, nz, colour, u, v, textured, normals)
	local position_x, position_y, position_z
	if scaling.mode == 1 then
		large_size = math.max(math.max(sprite.width, sprite.height), #sprite.layers)
		position_x = (x / large_size) - 0.5
		position_y = (y / large_size) - 0.5
		position_z = z / large_size
	elseif scaling.mode == 2 then
		position_x = (x - (sprite.width * 0.5)) * scaling.x
		position_y = (y - (sprite.height * 0.5)) * scaling.y
		position_z = z * scaling.z
	else
		position_x = x - (sprite.width * 0.5)
		position_y = y - (sprite.height * 0.5)
		position_z = z
	end
	local texture_u = ((colour % 16) / 16.0) + (u / 16.0)
	local texture_v = (math.floor(colour / 16) / 16.0) + (v / 16.0)
	if textured then
		for vertex_index = 1, #vertices do
			local vertex = vertices[vertex_index]
			if vertex.x == position_x and vertex.y == position_y and vertex.z == position_z and (not normals or (vertex.nx == nx and vertex.ny == ny and vertex.nz == nz)) and vertex.u == texture_u and vertex.v == texture_v then
				return vertex_index
			end
		end
	else
		for vertex_index = 1, #vertices do
			local vertex = vertices[vertex_index]
			if vertex.x == position_x and vertex.y == position_y and vertex.z == position_z and (not normals or (vertex.nx == nx and vertex.ny == ny and vertex.nz == nz)) and vertex.colour == colour then
				return vertex_index
			end
		end
	end
	if textured then
		vertices[#vertices + 1] = {
			x = position_x,
			y = position_y,
			z = position_z,
			nx = nx,
			ny = ny,
			nz = nz,
			u = texture_u,
			v = texture_v
		}
	else
		vertices[#vertices + 1] = {
			x = position_x,
			y = position_y,
			z = position_z,
			nx = nx,
			ny = ny,
			nz = nz,
			colour = colour
		}
	end
	return #vertices
end

--Outputs a structure containing a frame's mesh vertices, indices and a range of position values
local function build_mesh(sprite, layers, frame, scaling, textured, normals)
	local grid = { }
	for layer = 1, #layers do
		local cel = layers[layer]:cel(frame)
		local start = (cel ~= nil and Point(cel.bounds.x + 1, cel.bounds.y + 1) or Point(sprite.width, sprite.height))
		local finish = (cel ~= nil and Point(start.x + cel.bounds.width, start.y + cel.bounds.height) or Point(0, 0))
		image = (cel ~= nil and cel.image or nil)
		grid[layer] = { }
		for y = 1, sprite.height do
			grid[layer][y] = { }
			for x = 1, sprite.width do
				if x >= start.x and x < finish.x and y >= start.y and y < finish.y then
					grid[layer][y][x] = image:getPixel(x - start.x, y - start.y) & 0xFF
				else
					grid[layer][y][x] = 0
				end
			end
		end
	end
	local mesh_vertices = { }
	local mesh_indices = { }
	for layer = 1, #layers do
		for y = 1, sprite.height do
			for x = 1, sprite.width do
				local colour = grid[layer][y][x]
				if colour ~= 0 then
					if layer <= 1 or grid[layer - 1][y][x] == 0 then
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y - 1, layer - 1, scaling, 0.0, 0.0, -1.0, colour, 0.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y - 1, layer - 1, scaling, 0.0, 0.0, -1.0, colour, 1.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y, layer - 1, scaling, 0.0, 0.0, -1.0, colour, 1.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y - 1, layer - 1, scaling, 0.0, 0.0, -1.0, colour, 0.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y, layer - 1, scaling, 0.0, 0.0, -1.0, colour, 1.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y, layer - 1, scaling, 0.0, 0.0, -1.0, colour, 0.0, 0.0, textured, normals)
					end
					if layer >= #layers or grid[layer + 1][y][x] == 0 then
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y, layer, scaling, 0.0, 0.0, 1.0, colour, 1.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y - 1, layer, scaling, 0.0, 0.0, 1.0, colour, 1.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y - 1, layer, scaling, 0.0, 0.0, 1.0, colour, 0.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y, layer, scaling, 0.0, 0.0, 1.0, colour, 1.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y - 1, layer, scaling, 0.0, 0.0, 1.0, colour, 0.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y, layer, scaling, 0.0, 0.0, 1.0, colour, 0.0, 1.0, textured, normals)
					end
					if y <= 1 or grid[layer][y - 1][x] == 0 then
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y - 1, layer, scaling, 0.0, -1.0, 0.0, colour, 0.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y - 1, layer - 1, scaling, 0.0, -1.0, 0.0, colour, 0.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y - 1, layer - 1, scaling, 0.0, -1.0, 0.0, colour, 1.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y - 1, layer, scaling, 0.0, -1.0, 0.0, colour, 0.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y - 1, layer - 1, scaling, 0.0, -1.0, 0.0, colour, 1.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y - 1, layer, scaling, 0.0, -1.0, 0.0, colour, 1.0, 0.0, textured, normals)
					end
					if y >= sprite.height or grid[layer][y + 1][x] == 0 then
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y, layer - 1, scaling, 0.0, 1.0, 0.0, colour, 0.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y, layer - 1, scaling, 0.0, 1.0, 0.0, colour, 1.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y, layer, scaling, 0.0, 1.0, 0.0, colour, 1.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y, layer - 1, scaling, 0.0, 1.0, 0.0, colour, 0.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y, layer, scaling, 0.0, 1.0, 0.0, colour, 1.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y, layer, scaling, 0.0, 1.0, 0.0, colour, 0.0, 0.0, textured, normals)
					end
					if x <= 1 or grid[layer][y][x - 1] == 0 then
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y - 1, layer - 1, scaling, -1.0, 0.0, 0.0, colour, 0.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y, layer - 1, scaling, -1.0, 0.0, 0.0, colour, 1.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y, layer, scaling, -1.0, 0.0, 0.0, colour, 1.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y - 1, layer - 1, scaling, -1.0, 0.0, 0.0, colour, 0.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y, layer, scaling, -1.0, 0.0, 0.0, colour, 1.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x - 1, y - 1, layer, scaling, -1.0, 0.0, 0.0, colour, 0.0, 0.0, textured, normals)
					end
					if x >= sprite.width or grid[layer][y][x + 1] == 0 then
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y, layer, scaling, 1.0, 0.0, 0.0, colour, 0.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y, layer - 1, scaling, 1.0, 0.0, 0.0, colour, 0.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y - 1, layer - 1, scaling, 1.0, 0.0, 0.0, colour, 1.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y, layer, scaling, 1.0, 0.0, 0.0, colour, 0.0, 0.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y - 1, layer - 1, scaling, 1.0, 0.0, 0.0, colour, 1.0, 1.0, textured, normals)
						mesh_indices[#mesh_indices + 1] = add_vertex(sprite, mesh_vertices, x, y - 1, layer, scaling, 1.0, 0.0, 0.0, colour, 1.0, 0.0, textured, normals)
					end
				end
			end
		end
	end
	return {
		vertices = mesh_vertices,
		indices = mesh_indices,
		textured = textured,
		normals = normals
	}
end

--Outputs a binary string containing the sprite's palette as a PNG file
local function build_png(sprite, size)
	local full_width = 16 * size.width
	local full_height = 16 * size.height
	local palette = sprite.palettes[1]
	local palette_size = #palette
	local image = Image(full_width, full_height, ColorMode.RGB)
	for y = 1, full_height do
		local index = (math.tointeger((y - 1) / size.height) * 16) - 1
		for x = 1, full_width do
			if ((x - 1) % size.width) == 0 then index = index + 1 end
			image:drawPixel(x - 1, y - 1, palette:getColor(index < palette_size and index or 0))
		end
	end
	local path = app.fs.joinPath(app.fs.tempPath, "LayeredMesh3DPalette.png")
	image:saveAs(path)
	image = nil
	local file = io.open(path, "rb")
	local data = file:read("*all")
	file:close()
	return data
end

--Outputs a binary string containing the sprite's palette as a BMP file
local function build_bmp(sprite, size)
	local full_width = 16 * size.width
	local full_height = 16 * size.height
	local palette_size = #sprite.palettes[1]
	local file_header = "BM"..string_dword(54 + (palette_size * 4) + (full_width * full_height))..string_dword(0x0000)..string_dword(54 + (palette_size * 4))
	local info_header = string_dword(40)..string_dword(full_width)..string_dword(-full_height)..string_word(1)..string_word(8)..string_dword(0)..string_dword(0)..string_dword(0)..string_dword(0)..string_dword(palette_size < 256 and palette_size or 0)..string_dword(0)
	local palette = ""
	for index = 1, #sprite.palettes[1] do
		local colour = sprite.palettes[1]:getColor(index - 1)
		palette = palette..string_byte(colour.blue)..string_byte(colour.green)..string_byte(colour.red)..string_byte(0)
	end
	local pixels = ""
	for y = 1, full_height do
		local line = ""
		local index = (math.tointeger((y - 1) / size.height) * 16) - 1
		for x = 1, full_width do
			if ((x - 1) % size.width) == 0 then index = index + 1 end
			line = line..string_byte(index < palette_size and index or 0)
		end
		pixels = pixels..line
	end
	return file_header..info_header..palette..pixels
end

--Outputs a binary string containing the mesh indices and vertices while acquiring useful info
local function build_buffer(sprite, mesh, info)
	local data = ""
	for index = 1, #mesh.indices do
		data = data..string_word(mesh.indices[index] - 1)
	end
	for index = 1, #mesh.vertices do
		local vertex = mesh.vertices[index]
		info.min_x = math.min(info.min_x, vertex.x); info.max_x = math.max(info.max_x, vertex.x)
		info.min_y = math.min(info.min_y, vertex.y); info.max_y = math.max(info.max_y, vertex.y)
		info.min_z = math.min(info.min_z, vertex.z); info.max_z = math.max(info.max_z, vertex.z)
		info.min_nx = math.min(info.min_nx, vertex.nx); info.max_nx = math.max(info.max_nx, vertex.nx)
		info.min_ny = math.min(info.min_ny, vertex.ny); info.max_ny = math.max(info.max_ny, vertex.ny)
		info.min_nz = math.min(info.min_nz, vertex.nz); info.max_nz = math.max(info.max_nz, vertex.nz)
		data = data..string_float(vertex.x)..string_float(vertex.z)..string_float(vertex.y)..(mesh.normals and string_float(vertex.nx)..string_float(vertex.nz)..string_float(vertex.ny) or "")
		if mesh.textured then
			info.min_u = math.min(info.min_u, vertex.u); info.max_u = math.max(info.max_u, vertex.u)
			info.min_v = math.min(info.min_v, vertex.v); info.max_v = math.max(info.max_v, vertex.v)
			data = data..string_float(vertex.u)..string_float(vertex.v)
		else
			local colour = sprite.palettes[1]:getColor(vertex.colour)
			local red = ((1.0 / 256.0) * colour.red) ^ 2.4
			local green = ((1.0 / 256.0) * colour.green) ^ 2.4
			local blue = ((1.0 / 256.0) * colour.blue) ^ 2.4
			info.min_red = math.min(info.min_red, red); info.max_red = math.max(info.max_red, red)
			info.min_green = math.min(info.min_green, green); info.max_green = math.max(info.max_green, green)
			info.min_blue = math.min(info.min_blue, blue); info.max_blue = math.max(info.max_blue, blue)
			data = data..string_float(red)..string_float(green)..string_float(blue)
		end
	end
	return data
end

--Outputs an empty glTF file structure to be used when exporting
local function build_gltf_structure()
	return {
		scene = 0,
		scenes = { },
		nodes = { },
		meshes = { },
		materials = { },
		textures = { },
		samplers = { },
		images = { },
		buffers = { },
		views = { },
		accessors = { },
		add_scene = function(this, name, nodes)
			this.scenes[#this.scenes + 1] = {
				name = name,
				nodes = nodes
			}
			return #this.scenes
		end,
		add_node = function(this, mesh)
			this.nodes[#this.nodes + 1] = {
				mesh = mesh
			}
			return #this.nodes
		end,
		add_mesh = function(this, name, indices, positions, normals, colours_uvs, material, textured)
			this.meshes[#this.meshes + 1] = {
				name = name,
				indices = indices,
				positions = positions,
				normals = normals,
				colours_uvs = colours_uvs,
				material = material,
				textured = textured
			}
			return #this.meshes
		end,
		add_material = function(this, texture, metallic, roughness)
			this.materials[#this.materials + 1] = {
				texture = texture,
				metallic = metallic,
				roughness = roughness
			}
			return #this.materials
		end,
		add_texture = function(this, sampler, image)
			this.textures[#this.textures + 1] = {
				sampler = sampler,
				image = image
			}
			return #this.textures
		end,
		add_sampler = function(this, min, mag, s, t)
			this.samplers[#this.samplers + 1] = {
				min = min,
				mag = mag,
				s = s,
				t = t
			}
			return #this.samplers
		end,
		add_image = function(this, external, view, data)
			this.images[#this.images + 1] = {
				external = external,
				view = view,
				data = data
			}
			return #this.images
		end,
		add_buffer = function(this, data)
			this.buffers[#this.buffers + 1] = {
				data = data
			}
			return #this.buffers
		end,
		add_view = function(this, buffer, stride, offset, length, target)
			this.views[#this.views + 1] = {
				buffer = buffer,
				stride = stride,
				offset = offset,
				length = length,
				target = target
			}
			return #this.views
		end,
		add_accessor = function(this, view, offset, type, component, count, min, max)
			this.accessors[#this.accessors + 1] = {
				view = view,
				offset = offset,
				type = type,
				component = component,
				count = count,
				min = min,
				max = max
			}
			return #this.accessors
		end,
		output_binary = function(this)
			local header = "glTF"..string_dword(2)
			local json = "{\"scene\":"..this.scene..",\"scenes\":[{"
			local binary = ""
			for index, scene in ipairs(this.scenes) do
				json = json.."\"name\":\""..scene.name.."\",\"nodes\":["
				for node = 1, #scene.nodes do json = json..scene.nodes[node].."," end
				json = json:sub(1, #json - 1).."]},{"
			end
			json = json:sub(1, #json - 2).."],\"nodes\":[{"
			for index, node in ipairs(this.nodes) do
				json = json.."\"mesh\": "..node.mesh.."},{"
			end
			json = json:sub(1, #json - 2).."],\"meshes\":[{"
			for index, mesh in ipairs(this.meshes) do
				json = json.."\"name\":\""..mesh.name.."\",\"primitives\":[{\"indices\":"..mesh.indices..",\"attributes\":{\"POSITION\":"..mesh.positions..",\""..(mesh.normals >= 0 and"NORMAL\":"..mesh.normals..",\"" or "")..(mesh.textured and "TEXCOORD_0\":" or "COLOR_0\":")..mesh.colours_uvs.."},\"material\":"..mesh.material.."}]},{"
			end
			json = json:sub(1, #json - 2).."],\"materials\":[{"
			for index, material in ipairs(this.materials) do
				json = json.."\"pbrMetallicRoughness\":{\""..(material.texture >= 0 and "baseColorTexture\":{\"index\":"..material.texture.."}" or "baseColorFactor\":[1.0,1.0,1.0,1.0]")..",\"metallicFactor\":"..material.metallic..",\"roughnessFactor\":"..material.roughness.."}},{"
			end
			if #this.textures >= 1 then
				json = json:sub(1, #json - 2).."],\"textures\":[{"
				for index, texture in ipairs(this.textures) do
					json = json.."\"source\":"..texture.image..",\"sampler\":"..texture.sampler.."},{"
				end
				json = json:sub(1, #json - 2).."],\"samplers\":[{"
				for index, sampler in ipairs(this.samplers) do
					json = json.."\"minFilter\":"..sampler.min..",\"magFilter\":"..sampler.mag..",\"wrapS\":"..sampler.s..",\"wrapT\":"..sampler.t.."},{"
				end
				json = json:sub(1, #json - 2).."],\"images\":[{"
				for index, image in ipairs(this.images) do
					json = json.."\"bufferView\":"..image.view..",\"mimeType\":\"image/png\"},{"
				end
			end
			json = json:sub(1, #json - 2).."],\"buffers\":[{"
			for index, buffer in ipairs(this.buffers) do
				json = json.."\"byteLength\":"..#buffer.data.."},{"
				binary = binary..buffer.data
			end
			json = json:sub(1, #json - 2).."],\"bufferViews\":[{"
			for index, view in ipairs(this.views) do
				json = json.."\"buffer\":"..view.buffer..","..(view.stride >= 1 and "\"byteStride\":"..view.stride.."," or "").."\"byteOffset\":"..view.offset..",\"byteLength\":"..view.length..(view.target >= 0 and ",\"target\":"..view.target or "").."},{"
			end
			json = json:sub(1, #json - 2).."],\"accessors\":[{"
			for index, accessor in ipairs(this.accessors) do
				json = json.."\"bufferView\":"..accessor.view..",\"byteOffset\":"..accessor.offset..",\"type\":\""..accessor.type.."\",\"componentType\":"..accessor.component..",\"count\":"..accessor.count..",\"min\":["
				for value = 1, #accessor.min do
					json = json..accessor.min[value]..","
				end
				json = json:sub(1, #json - 1).."],\"max\":["
				for value = 1, #accessor.max do
					json = json..accessor.max[value]..","
				end
				json = json:sub(1, #json - 1).."]},{"
			end
			json = json:sub(1, #json - 2).." ],\"asset\":{\"version\":\"2.0\",\"minVersion\":\"2.0\",\"generator\":\"LayeredMesh3D For Asperite\"}}"
			json = json..(({ "", "   ", "  ", " " })[(#json % 4) + 1])
			json = string_dword(#json).."JSON"..json
			binary = binary..(({ "", "\0\0\0", "\0\0", "\0" })[(#binary % 4) + 1])
			binary = string_dword(#binary).."BIN\0"..binary
			header = header..string_dword(#header + #json + #binary + 4)
			return header..json..binary
		end,
		output_text = function(this)
			local data = "{\n\t\"scene\": "..this.scene..",\n\t\"scenes\": [ {"
			for index, scene in ipairs(this.scenes) do
				data = data.."\n\t\t\"name\": \""..scene.name.."\",\n\t\t\"nodes\": [ "
				for node = 1, #scene.nodes do data = data..scene.nodes[node]..", " end
				data = data:sub(1, #data - 2).." ]\n\t}, {"
			end
			data = data:sub(1, #data - 3).." ],\n\t\"nodes\": [ {"
			for index, node in ipairs(this.nodes) do
				data = data.."\n\t\t\"mesh\": "..node.mesh.."\n\t}, {"
			end
			data = data:sub(1, #data - 3).." ],\n\t\"meshes\": [ {"
			for index, mesh in ipairs(this.meshes) do
				data = data.."\n\t\t\"name\": \""..mesh.name.."\",\n\t\t\"primitives\": [ {\n\t\t\t\"indices\": "..mesh.indices..",\n\t\t\t\"attributes\": {\n\t\t\t\t\"POSITION\": "..mesh.positions..",\n\t\t\t\t\""..(mesh.normals >= 0 and"NORMAL\": "..mesh.normals..",\n\t\t\t\t\"" or "")..(mesh.textured and "TEXCOORD_0\": " or "COLOR_0\": ")..mesh.colours_uvs.."\n\t\t\t},\n\t\t\t\"material\": "..mesh.material.."\n\t\t} ]\n\t}, {"
			end
			data = data:sub(1, #data - 3).." ],\n\t\"materials\": [ {"
			for index, material in ipairs(this.materials) do
				data = data.."\n\t\t\"pbrMetallicRoughness\": {\n\t\t\t\""..(material.texture >= 0 and "baseColorTexture\": {\n\t\t\t\t\"index\": "..material.texture.."\n\t\t\t}" or "baseColorFactor\": [ 1.0, 1.0, 1.0, 1.0 ]")..",\n\t\t\t\"metallicFactor\": "..material.metallic..",\n\t\t\t\"roughnessFactor\": "..material.roughness.."\n\t\t}\n\t}, {"
			end
			if #this.textures >= 1 then
				data = data:sub(1, #data - 3).." ],\n\t\"textures\": [ {"
				for index, texture in ipairs(this.textures) do
					data = data.."\n\t\t\"source\": "..texture.image..",\n\t\t\"sampler\": "..texture.sampler.."\n\t}, {"
				end
				data = data:sub(1, #data - 3).." ],\n\t\"samplers\": [ {"
				for index, sampler in ipairs(this.samplers) do
					data = data.."\n\t\t\"minFilter\": "..sampler.min..",\n\t\t\"magFilter\": "..sampler.mag..",\n\t\t\"wrapS\": "..sampler.s..",\n\t\t\"wrapT\": "..sampler.t.."\n\t}, {"
				end
				data = data:sub(1, #data - 3).." ],\n\t\"images\": [ {"
				for index, image in ipairs(this.images) do
					data = data.."\n\t\t\""..(image.view >= 0 and "bufferView\": "..image.view or "uri\": \""..(image.external and image.data or "data:image/png;base64,"..string_base64(image.data)).."\"").."\n\t}, {"
				end
			end
			data = data:sub(1, #data - 3).." ],\n\t\"buffers\": [ {"
			for index, buffer in ipairs(this.buffers) do
				data = data.."\n\t\t\"uri\": \"data:application/gltf-buffer;base64,"..string_base64(buffer.data).."\",\n\t\t\"byteLength\": "..#buffer.data.."\n\t}, {"
			end
			data = data:sub(1, #data - 3).." ],\n\t\"bufferViews\": [ {"
			for index, view in ipairs(this.views) do
				data = data.."\n\t\t\"buffer\": "..view.buffer..",\n\t\t"..(view.stride >= 1 and "\"byteStride\": "..view.stride..",\n\t\t" or "").."\"byteOffset\": "..view.offset..",\n\t\t\"byteLength\": "..view.length..",\n\t\t\"target\": "..view.target.."\n\t}, {"
			end
			data = data:sub(1, #data - 3).." ],\n\t\"accessors\": [ {"
			for index, accessor in ipairs(this.accessors) do
				data = data.."\n\t\t\"bufferView\": "..accessor.view..",\n\t\t\"byteOffset\": "..accessor.offset..",\n\t\t\"type\": \""..accessor.type.."\",\n\t\t\"componentType\": "..accessor.component..",\n\t\t\"count\": "..accessor.count..",\n\t\t\"min\": [ "
				for value = 1, #accessor.min do
					data = data..accessor.min[value]..", "
				end
				data = data:sub(1, #data - 2).." ],\n\t\t\"max\": [ "
				for value = 1, #accessor.max do
					data = data..accessor.max[value]..", "
				end
				data = data:sub(1, #data - 2).." ]\n\t}, {"
			end
			return data:sub(1, #data - 3).." ],\n\t\"asset\": {\n\t\t\"version\": \"2.0\",\n\t\t\"minVersion\": \"2.0\",\n\t\t\"generator\": \"LayeredMesh3D For Asperite\"\n\t}\n}"
		end
	}
end

local function export_glb(sprite, layers, name, from, to, scaling, mesh_path, texture_size, textured, normals)
	local gltf = build_gltf_structure()
	local image = ""
	local offset = 0
	if textured then
		offset = 1
		image = build_png(sprite, texture_size)
		gltf:add_view(0, 0, 0, #image, -1)
	end
	local accessor_count = (normals and 4 or 3)
	local style_offset = 12 + (normals and 12 or 0)
	local vertex_size = style_offset + (textured and 8 or 12)
	local added_scenes = { }
	local added_buffers = { }
	local buffer_data = image..(({ "", "\0\0\0", "\0\0", "\0" })[(#image % 4) + 1])
	local frame_range = to - from
	for frame = 0, frame_range do
		added_scenes[#added_scenes + 1] = frame
		local info = {
			min_x = 1000000.0, max_x = -1000000.0,
			min_y = 1000000.0, max_y = -1000000.0,
			min_z = 1000000.0, max_z = 0.0,
			min_nx = 1000000.0, max_nx = -1000000.0,
			min_ny = 1000000.0, max_ny = -1000000.0,
			min_nz = 1000000.0, max_nz = -1000000.0,
			min_red = 1.0, max_red = 0.0,
			min_green = 1.0, max_green = 0.0,
			min_blue = 1.0, max_blue = 0.0,
			min_u = 1.0, max_u = 0.0,
			min_v = 1.0, max_v = 0.0
		}
		local mesh = build_mesh(sprite, layers, from + frame, scaling, textured, normals)
		local buffer = build_buffer(sprite, mesh, info)
		local buffer_index = -1
		for added_index, added_data in ipairs(added_buffers) do
			if added_data == buffer then
				buffer_index = added_index - 1
				break
			end
		end
		local new_buffer = false
		if buffer_index <= -1 then
			added_buffers[#added_buffers + 1] = buffer
			buffer_index = #added_buffers - 1
			new_buffer = true
		end
		gltf:add_node(frame)
		gltf:add_mesh(name..(frame_range > 0 and frame or ""), buffer_index * accessor_count, (buffer_index * accessor_count) + 1, (normals and (buffer_index * accessor_count) + 2 or -1), (buffer_index * accessor_count) + (normals and 3 or 2), 0, textured)
		if new_buffer then
			local index_length = #mesh.indices * 2
			local vertex_length = #mesh.vertices * vertex_size
			gltf:add_view(0, 0, #buffer_data, index_length, 34963)
			gltf:add_view(0, vertex_size, #buffer_data + index_length, vertex_length, 34962)
			gltf:add_accessor(offset + buffer_index * 2, 0, "SCALAR", 5123, #mesh.indices, { 0 }, { #mesh.vertices - 1 })
			gltf:add_accessor(offset + (buffer_index * 2) + 1, 0, "VEC3", 5126, #mesh.vertices, { info.min_x, info.min_z, info.min_y }, { info.max_x, info.max_z, info.max_y })
			if normals then
				gltf:add_accessor(offset +(buffer_index * 2) + 1, 12, "VEC3", 5126, #mesh.vertices, { info.min_nx, info.min_nz, info.min_ny }, { info.max_nx, info.max_nz, info.max_ny })
			end
			gltf:add_accessor(offset + (buffer_index * 2) + 1, style_offset, (textured and "VEC2" or "VEC3"), 5126, #mesh.vertices, (textured and { info.min_u, info.min_v } or { info.min_red, info.min_green, info.min_blue }), (textured and { info.max_u, info.max_v } or { info.max_red, info.max_green, info.max_blue }))
			buffer_data = buffer_data..buffer
		end
	end
	gltf:add_scene(name, added_scenes)
	if textured then
		gltf:add_material(0, 0.0, 1.0)
		gltf:add_texture(0, 0)
		gltf:add_sampler(9728, 9728, 33071, 33071)
		gltf:add_image(false, 0, nil)
	else
		gltf:add_material(-1, 0.0, 1.0)
	end
	gltf:add_buffer(buffer_data)
	local file = io.open(mesh_path, "wb")
	file:write(gltf:output_binary())
	file:close()
	return true
end

--Exports the glTF mesh in a text format to the specified location
local function export_gltf(sprite, layers, name, from, to, scaling, mesh_path, texture_path, texture_size, textured, normals)
	local gltf = build_gltf_structure()
	local accessor_count = (normals and 4 or 3)
	local style_offset = 12 + (normals and 12 or 0)
	local vertex_size = style_offset + (textured and 8 or 12)
	local added_scenes = { }
	local added_buffers = { }
	local frame_range = to - from
	for frame = 0, frame_range do
		added_scenes[#added_scenes + 1] = frame
		local info = {
			min_x = 1000000.0, max_x = -1000000.0,
			min_y = 1000000.0, max_y = -1000000.0,
			min_z = 1000000.0, max_z = 0.0,
			min_nx = 1000000.0, max_nx = -1000000.0,
			min_ny = 1000000.0, max_ny = -1000000.0,
			min_nz = 1000000.0, max_nz = -1000000.0,
			min_red = 1.0, max_red = 0.0,
			min_green = 1.0, max_green = 0.0,
			min_blue = 1.0, max_blue = 0.0,
			min_u = 1.0, max_u = 0.0,
			min_v = 1.0, max_v = 0.0
		}
		local mesh = build_mesh(sprite, layers, from + frame, scaling, textured, normals)
		local buffer = build_buffer(sprite, mesh, info)
		local buffer_index = -1
		for added_index, added_data in ipairs(added_buffers) do
			if added_data == buffer then
				buffer_index = added_index - 1
				break
			end
		end
		local new_buffer = false
		if buffer_index <= -1 then
			added_buffers[#added_buffers + 1] = buffer
			buffer_index = #added_buffers - 1
			new_buffer = true
		end
		gltf:add_node(frame)
		gltf:add_mesh(name..(frame_range > 0 and frame or ""), buffer_index * accessor_count, (buffer_index * accessor_count) + 1, (normals and (buffer_index * accessor_count) + 2 or -1), (buffer_index * accessor_count) + (normals and 3 or 2), 0, textured)
		if new_buffer then
			local index_length = #mesh.indices * 2
			local vertex_length = #mesh.vertices * vertex_size
			gltf:add_buffer(buffer)
			gltf:add_view(buffer_index, 0, 0, index_length, 34963)
			gltf:add_view(buffer_index, vertex_size, index_length, vertex_length, 34962)
			gltf:add_accessor(buffer_index * 2, 0, "SCALAR", 5123, #mesh.indices, { 0 }, { #mesh.vertices - 1 })
			gltf:add_accessor((buffer_index * 2) + 1, 0, "VEC3", 5126, #mesh.vertices, { info.min_x, info.min_z, info.min_y }, { info.max_x, info.max_z, info.max_y })
			if normals then
				gltf:add_accessor((buffer_index * 2) + 1, 12, "VEC3", 5126, #mesh.vertices, { info.min_nx, info.min_nz, info.min_ny }, { info.max_nx, info.max_nz, info.max_ny })
			end
			gltf:add_accessor((buffer_index * 2) + 1, style_offset, (textured and "VEC2" or "VEC3"), 5126, #mesh.vertices, (textured and { info.min_u, info.min_v } or { info.min_red, info.min_green, info.min_blue }), (textured and { info.max_u, info.max_v } or { info.max_red, info.max_green, info.max_blue }))
		end
	end
	gltf:add_scene(name, added_scenes)
	if textured then
		gltf:add_material(0, 0.0, 1.0)
		gltf:add_texture(0, 0)
		gltf:add_sampler(9728, 9728, 33071, 33071)
		if texture_path ~= nil then
			gltf:add_image(true, -1, texture_path)
		else
			gltf:add_image(false, -1, build_png(sprite, texture_size))
		end
	else
		gltf:add_material(-1, 0.0, 1.0)
	end
	local file = io.open(mesh_path, "w")
	file:write(gltf:output_text())
	file:close()
	return true
end

--Exports the mesh to the specified location
local function export_mesh(sprite, layers, name, from, to, scaling, mesh_path, texture_path, texture_size, textured, normals)
	if mesh_path == nil or mesh_path == "" then
		app.alert{ title = "Error", text = "No mesh file was specified" }
		return false
	end
	local frame_from = from
	local frame_to = to
	if from > to then
		frame_from = to
		frame_to = from
	end
	frame_from = math.max(frame_from, 1)
	frame_to = math.min(frame_to, #sprite.frames)
	local extension = string.upper(app.fs.fileExtension(mesh_path))
	if extension == "GLB" then return export_glb(sprite, layers, name, frame_from, frame_to, scaling, mesh_path, texture_size, textured, normals)
	elseif extension == "GLTF" then return export_gltf(sprite, layers, name, frame_from, frame_to, scaling, mesh_path, texture_path, texture_size, textured, normals)
	else
		app.alert{ title = "Error", text = "Specified mesh format isn't supported" }
		return false
	end
end

--Exports the texture to the specified location
local function export_texture(sprite, path, size)
	if path == nil then return end
	local data
	local extension = string.upper(app.fs.fileExtension(path))
	if extension == "PNG" then data = build_png(sprite, size)
	elseif extension == "BMP" then data = build_bmp(sprite, size)
	else app.alert{ title = "Error", text = "Specified texture format isn't supported" } end
	local file = io.open(path, "wb")
	file:write(data)
	file:close()
end


--					SCRIPT					--

--Acquire the needed sprite and layers
local sprite = acquire_sprite()
if sprite == nil then return end
local layers = acquire_layers(sprite)
if #layers == 0 then
	app.alert{ title = "Error", text = "Sprite must have at least 1 image layer" }
	return
end

--Create, set up and show the custom dialog then catch the results
local dialog = create_dialog(sprite, sprite.filename:gsub("."..app.fs.fileExtension(sprite.filename), ""))
dialog:show{ wait = true }
local dialog_data = dialog.data
if not dialog_data.confirm then return end

--Export the mesh and texture if the option was enabled
local texture_size = Size(dialog_data.texture_width, dialog_data.texture_height)
local texture_exported = dialog_data.format_text and dialog_data.is_textured and not dialog_data.embedded_texture
if not export_mesh(sprite, layers, dialog_data.name_entry, math.tointeger(dialog_data.frame_from), math.tointeger(dialog_data.frame_to), { x = dialog_data.scale_x, y = dialog_data.scale_y, z = dialog_data.scale_z, mode = ((dialog_data.no_scaling) and 0 or (dialog_data.normalised_scaling and 1 or 2)) }, dialog_data.mesh_file, (texture_exported and app.fs.fileName(dialog_data.texture_file) or nil), texture_size, dialog_data.is_textured, dialog_data.flat_normals) then return end
if texture_exported then export_texture(sprite, dialog_data.texture_file, texture_size) end

