--					GLOBALS					--

local dialog
local sprite = app.sprite

local face_width = sprite ~= nil and sprite.width or 0
local face_height = sprite ~= nil and sprite.height or 0
local face_depth = sprite ~= nil and #sprite.layers or 0

local slice_top = 1
local slice_bottom = 1
local slice_front = 1
local slice_back = 1
local slice_left = 1
local slice_right = 1
local slice_active = 1

--					FUNCTIONS					--

--Redraws the dialog's viewport canvas
local function redraw_dialog() dialog:repaint() end

--Changes the face being viewed in the viewport
local function change_face()
	if sprite ~= nil then
	local dialog_data = dialog.data
		if dialog_data.top_face or dialog_data.bottom_face then
			face_width = sprite.width
			face_height = sprite.height
			face_depth = #sprite.layers
			slice_active = (dialog_data.top_face and slice_top or slice_bottom)
		elseif dialog_data.front_face or dialog_data.back_face then
			face_width = sprite.width
			face_height = #sprite.layers
			face_depth = sprite.height
			slice_active = (dialog_data.front_face and slice_front or slice_back)
		elseif dialog_data.left_face or dialog_data.right_face then
			face_width = sprite.height
			face_height = #sprite.layers
			face_depth = sprite.width
			slice_active = (dialog_data.left_face and slice_left or slice_right)
		end
		slice_active = math.min(math.max(slice_active, 1), face_depth)
		dialog:modify{
			id = "slice",
			text = ""..slice_active
		}
	end
	dialog:repaint()
end

--Swaps the active sprite's redraw functions
local function sprite_changed()
	if app.sprite ~= sprite then
		if sprite ~= nil then sprite.events:off(redraw_dialog) end
		sprite = app.sprite
		if sprite ~= nil then sprite.events:on(redraw_dialog) end
	end
	change_face()
end

local function slice_changed()
	local dialog_data = dialog.data
	slice_active = math.min(math.max(dialog_data.slice, 1), face_depth)
	dialog:modify{
		id = "slice",
		text = ""..slice_active
	}
	if dialog_data.top_face then slice_top = slice_active
	elseif dialog_data.bottom_face then slice_bottom = slice_active
	elseif dialog_data.front_face then slice_front = slice_active
	elseif dialog_data.back_face then slice_back = slice_active
	elseif dialog_data.front_face then slice_front = slice_active
	elseif dialog_data.left_face then slice_left = slice_active
	else slice_right = slice_active end
	dialog:repaint()
end

--Creates and sets up the viewport dialog
local function create_dialog()
	dialog = Dialog{
		title = "Layered Mesh 3D Viewport",
		onclose = function()
			sprite.events:off(redraw_dialog)
			app.events:off(sprite_changed)
		 end
	}
	dialog:canvas{
		id = "viewport",
		width = 256,
		height = 256,
		onpaint = function(event)
			if sprite == nil then return end
			local dialog_data = dialog.data
			local layer_count = #sprite.layers
			local image = Image(face_width, face_height, ColorMode.RGB)
			for y = 1, face_height do
				for x = 1, face_width do
					local colour = 0
					local depth = 1.0
					for z = 1, face_depth do
						if dialog_data.show_all or z == slice_active then
							if dialog_data.top_face then
								local layer = sprite.layers[layer_count - (z - 1)]
								local cel = ((layer ~= nil and layer.isImage) and layer:cel(app.frame) or nil)
								if cel ~= nil then 
									local bounds = cel.bounds
									if x > bounds.x and y > bounds.y and x <= (bounds.x + bounds.width) and y <= (bounds.y + bounds.height) then
										colour = cel.image:getPixel(x - (bounds.x + 1), y - (bounds.y + 1))
										depth = 1.0 - ((0.5 / face_depth) * (z - 1))
									end
								end
							elseif dialog_data.bottom_face then
								local layer = sprite.layers[z]
								local cel = ((layer ~= nil and layer.isImage) and layer:cel(app.frame) or nil)
								if cel ~= nil then
									local bounds = cel.bounds
									local flipped = face_width - (x - 1)
									if flipped > bounds.x and y > bounds.y and flipped <= (bounds.x + bounds.width) and y <= (bounds.y + bounds.height) then
										colour = cel.image:getPixel(flipped - (bounds.x + 1), y - (bounds.y + 1))
										depth = 1.0 - ((0.5 / face_depth) * (z - 1))
									end
								end
							elseif dialog_data.front_face then
								local layer = sprite.layers[layer_count - (y - 1)]
								local cel = ((layer ~= nil and layer.isImage) and layer:cel(app.frame) or nil)
								if cel ~= nil then
									local bounds = cel.bounds
									if x > bounds.x and z > bounds.y and x <= (bounds.x + bounds.width) and z <= (bounds.y + bounds.height) then
										colour = cel.image:getPixel(x - (bounds.x + 1), (bounds.y + bounds.height) - z)
										depth = 1.0 - ((0.5 / face_depth) * (z - 1))
									end
								end
							elseif dialog_data.back_face then
								local layer = sprite.layers[layer_count - (y - 1)]
								local cel = ((layer ~= nil and layer.isImage) and layer:cel(app.frame) or nil)
								if cel ~= nil then
									local bounds = cel.bounds
									local flipped = face_width - (x - 1)
									if flipped > bounds.x and z > bounds.y and flipped <= (bounds.x + bounds.width) and z <= (bounds.y + bounds.height) then
										colour = cel.image:getPixel(flipped - (bounds.x + 1), z - (bounds.y + 1))
										depth = 1.0 - ((0.5 / face_depth) * (z - 1))
									end
								end
							elseif dialog_data.left_face then
								local layer = sprite.layers[layer_count - (y - 1)]
								local cel = ((layer ~= nil and layer.isImage) and layer:cel(app.frame) or nil)
								if cel ~= nil then
									local bounds = cel.bounds
									if z > bounds.x and x > bounds.y and z <= (bounds.x + bounds.width) and x <= (bounds.y + bounds.height) then
										colour = cel.image:getPixel(z - (bounds.x + 1), x - (bounds.y + 1))
										depth = 1.0 - ((0.5 / face_depth) * (z - 1))
									end
								end
							elseif dialog_data.right_face then
								local layer = sprite.layers[layer_count - (y - 1)]
								local cel = ((layer ~= nil and layer.isImage) and layer:cel(app.frame) or nil)
								if cel ~= nil then
									local bounds = cel.bounds
									local flipped = face_width - (x - 1)
									if z > bounds.x and flipped > bounds.y and z <= (bounds.x + bounds.width) and flipped <= (bounds.y + bounds.height) then
										colour = cel.image:getPixel(bounds.width - (z - bounds.x), flipped - (bounds.y + 1))
										depth = 1.0 - ((0.5 / face_depth) * (z - 1))
									end
								end
							end
							if colour ~= 0 then
								break
							end
						end
					end
					if colour ~= 0 then
						colour = sprite.palettes[1]:getColor(colour)
						if dialog_data.show_colour and dialog_data.show_depth then image:drawPixel(x - 1, y - 1, Color(colour.red * depth, colour.green * depth, colour.blue * depth))
						elseif dialog_data.show_colour and not dialog_data.show_depth then image:drawPixel(x - 1, y - 1, colour)
						elseif not dialog_data.show_colour and dialog_data.show_depth then image:drawPixel(x - 1, y - 1, Color(255 * depth, 255 * depth, 255 * depth)) end
					end
				end
			end
			local graphics = event.context
			local aspect_x = graphics.width / face_width
			local aspect_y = graphics.height / face_height
			local draw_aspect = math.min(aspect_x, aspect_y) * (dialog_data.zoom / 100.0)
			local draw_width = face_width * draw_aspect
			local draw_height = face_height * draw_aspect
			graphics:drawImage(image, Rectangle(0, 0, face_width, face_height), Rectangle((graphics.width * 0.5) - (draw_width * 0.5), (graphics.height * 0.5) - (draw_height * 0.5), draw_width, draw_height))
		end
	}
	dialog:separator{ }
	dialog:radio{
		id = "top_face",
		label = "Face",
		text = "Top",
		selected = true,
		onclick = change_face
	}
	dialog:radio{
		id = "bottom_face",
		text = "Bottom",
		selected = false,
		onclick = change_face
	}
	dialog:radio{
		id = "front_face",
		text = "Front",
		selected = false,
		onclick = change_face
	}
	dialog:radio{
		id = "back_face",
		text = "Back",
		selected = false,
		onclick = change_face
	}
	dialog:radio{
		id = "left_face",
		text = "Left",
		selected = false,
		onclick = change_face
	}
	dialog:radio{
		id = "right_face",
		text = "Right",
		selected = false,
		onclick = change_face
	}
	dialog:check{
		id = "show_all",
		label = "Show",
		text = "All Slices",
		selected = true,
		onclick = redraw_dialog
	}
	dialog:check{
		id = "show_colour",
		text = "Colour",
		selected = true,
		onclick = redraw_dialog
	}
	dialog:check{
		id = "show_depth",
		text = "Depth",
		selected = true,
		onclick = redraw_dialog
	}
	dialog:number{
		id = "slice",
		label = "Slice",
		min = "1",
		max = ""..face_depth,
		value = "1",
		onchange = slice_changed
	}
	dialog:slider{
		id = "zoom",
		label = "Zoom",
		min = 1,
		max = 100,
		value = 50,
		onchange = redraw_dialog
	}
	change_face()
	return dialog
end


--					SCRIPT					--

if sprite ~= nil then sprite.events:on("change", redraw_dialog) end
app.events:on("sitechange", sprite_changed)
create_dialog():show{ wait = false }

